# -*- coding: utf-8 -*-
import sys
import time

from blindspin import spinner

import click

import emoji

from . import test
from .. import cli, options
from ..api import Apps, Config, Releases


@cli.cli.command()
@click.option('--message', help='Deployment message')
@click.option('--hard', is_flag=True,
              help='Pull the latest service images on deploy')
@options.app(allow_option=False)
def deploy(app, message, hard):
    """Deploy your app to Storyscript Cloud."""
    cli.user()

    payload = test.compile_app(app, False)  # Also adds a spinner.

    if payload is None:
        sys.exit(1)  # Error already printed by compile_app.

    click.echo(f'Deploying app {app}... ', nl=False)

    with spinner():
        config = Config.get(app)
        release = Releases.create(config, payload, app, message, hard)

    url = f'https://{app}.storyscriptapp.com/'
    click.echo()
    click.echo(
        click.style('\b' + emoji.emojize(':heavy_check_mark:'), fg='green')
        + f' Version {release["id"]} of your app has '
        f'been queued for deployment.\n'
    )

    click.echo('Waiting for deployment to complete…  ', nl=False)
    with spinner():
        if Apps.maintenance(app, maintenance=None):
            click.echo()
            click.echo()
            click.echo(
                'Your app is in maintenance mode.\n'
                'Run the following to turn off it off:'
            )
            cli.print_command('story maintenance off')
            click.echo()
            click.echo(
                'Once maintenance mode is turned off, '
                'your app will be deployed immediately.'
            )
            return

        state = 'QUEUED'
        while state in ['DEPLOYING', 'QUEUED']:
            state = Releases.get(app)[0]['state']
            time.sleep(0.5)

    click.echo()
    if state == 'DEPLOYED':
        click.echo(
            click.style('\b' + emoji.emojize(':heavy_check_mark:'), fg='green')
            + ' Deployment successful!'
        )
        click.echo(
            f'If your Story responds to HTTP requests, please visit:\n  {url}'
        )
    elif state == 'FAILED':
        click.echo(
            click.style('X', fg='red') + ' Deployment failed!', err=True
        )
        click.echo(
            'Please use the following command to view your app\'s logs:',
            err=True,
        )
        cli.print_command('story logs')
    else:
        click.echo(
            f'An unhandled state of your app has been encountered - {state}',
            err=True,
        )
        click.echo(f'Please shoot an email to support@storyscript.io')



# -*- coding: utf-8 -*-

import click

from .. import cli


@cli.cli.command(hidden=True)
def update():
    """Look for new version updates to CLI."""
    # TODO create update command
    click.echo('Sorry, command not programmed yet.')



import json
import os
from unittest.mock import MagicMock

from click.testing import CliRunner, Result

import pytest

STORYSCRIPT_CONFIG = {
    'id': os.environ.get('STORYSCRIPT_INT_CONF_USER_ID', 'my_user_id'),
    'access_token': os.environ.get('STORYSCRIPT_INT_CONF_ACCESS_TOKEN',
                                   'my_access_token'),
    'email': 'me@me.com'
}


@pytest.fixture
def pre_init_cli_runner():
    def wrapper(init_with_config_path='config.json',
                write_default_config=True):
        if write_default_config:
            with open(init_with_config_path, 'w') as f:
                f.write(json.dumps(STORYSCRIPT_CONFIG))
        from story import cli
        cli.init(config_path=init_with_config_path)

    return wrapper


@pytest.fixture
def runner(magic, pre_init_cli_runner):
    cli_runner = CliRunner()

    def function(command_function, stdin: str = None,
                 init_with_config_path='config.json',
                 write_default_config=True,
                 exit_code: int = 0,
                 args: list = []) -> Result:
        pre_init_cli_runner(init_with_config_path, write_default_config)

        result = cli_runner.invoke(
            command_function,
            input=stdin,
            args=args)

        if result.exception is not None \
                and not isinstance(result.exception, SystemExit):
            print(result.exc_info)
            print(result.stderr)
            assert False

        assert result.exit_code == exit_code, \
            f'stdout: {result.stdout}\n\nstderr: {result.stderr}'
        return result

    out = magic()
    out.run = function
    out.runner = cli_runner
    return out


@pytest.fixture
def magic(mocker):
    """
    Shorthand for mocker.MagicMock. It's magic!
    """
    return mocker.MagicMock


@pytest.fixture
def patch_init(mocker):
    """
    Makes patching a class' constructor slightly easier
    """
    def patch_init(item):
        mocker.patch.object(item, '__init__', return_value=None)
    return patch_init


@pytest.fixture
def patch_many(mocker):
    """
    Makes patching many attributes of the same object simpler
    """
    def patch_many(item, attributes):
        for attribute in attributes:
            mocker.patch.object(item, attribute)
    return patch_many


@pytest.fixture
def patch(mocker, patch_init, patch_many):
    mocker.patch.init = patch_init
    mocker.patch.many = patch_many
    return mocker.patch


@pytest.fixture
def async_mock():
    def return_value(*args, **kwargs):
        """
        Inspired from
        https://blog.miguelgrinberg.com/post/unit-testing-asyncio-code.
        """
        m = MagicMock(*args, **kwargs)

        async def mock_coro(*args, **kwargs):
            return m(*args, **kwargs)

        mock_coro.mock = m
        return mock_coro

    return return_value


@pytest.fixture
def init_sample_app_in_cwd():
    def cb():
        with open('story.yml', 'w') as f:
            f.write('app_name: my_app\n')

        os.mkdir('src')
        os.mkdir('src/b')
        with open('src/a.story', 'w') as f:
            f.write('a = 1 + 0\n')

        with open('src/b/b.story', 'w') as f:
            f.write('b = 0 + 1\n')

    return cb


@pytest.fixture
def app_dir():
    pass



# -*- coding: utf-8 -*-
import os
import pkgutil
import sys

import click

import emoji

from story import cli


def parse_story(story_path):
    with open(story_path, 'r') as f:
        contents = f.read()

    first = contents.split('\n')[0]
    return ''.join(first.split('#')[1:]).strip()


def get_stories():
    d = os.path.dirname(sys.modules['story'].__file__)
    d = os.path.join(d, 'stories')

    collection = []
    for filename in os.listdir(d):
        path = os.path.join(d, filename)
        story = {
            'name': filename.split('.')[0],
            'filename': filename,
            'path': path,
            'dir': d,
            'description': parse_story(path),
        }
        collection.append(story)

    return collection


STORIES = get_stories()
CHOICES = [story['name'] for story in STORIES]
CHOICES += ['-']


@cli.cli.command()
@click.argument('story', default='-', type=click.Choice(CHOICES))
@click.argument(
    'output_file', default=None, type=click.Path(exists=False), required=False
)
def write(story, output_file=None):
    """Pre–defined Storyscripts for your app!"""

    if story == '-':
        click.echo(click.style('Please specify a template:', bold=True))

        # Print the stories to the console.
        for story in STORIES:
            part1 = f"  {story['name']}"
            part2 = (' ' * (13 - len(part1))) + f'- {story["description"]}'

            click.echo(click.style(part1, fg='cyan') + part2)
        click.echo()
        click.echo(
            '  Run $ '
            + click.style('story write :template_name: ', fg='magenta')
            + emoji.emojize(':backhand_index_pointing_left:')
        )
        click.echo('')

        click.echo(click.style('Learn more:', bold=True))

        click.echo(
            '  - Services: '
            + click.style('https://hub.storyscript.io/', fg='cyan')
        )

        click.echo(
            '  - Examples: '
            + click.style(
                'https://github.com/storyscript/examples', fg='cyan'
            )
        )

        # click.echo('')
        sys.exit(1)

    else:

        # Grab the story, from packaging...
        data = pkgutil.get_data('story', f'stories/{story}.story')

        # If output_file was passed, assume it was an interfactive session.
        if output_file:
            # Write to the file...
            with open(output_file, 'wb') as f:
                f.write(data)

            cmd = f'cat {output_file}'
            cmd = click.style(cmd, fg='magenta')
            click.echo(f'$ {cmd}', err=True)

        click.echo(data)

        app_name = cli.get_app_name_from_yml()
        if app_name is None:
            app_name = 'Not created yet'

        cli.track(
            'App Bootstrapped', {'App name': app_name, 'Template used': story}
        )
        sys.exit(0)



# -*- coding: utf-8 -*-

import click

from .. import cli


@cli.cli.command()
def login():
    """Login to Storyscript Cloud via GitHub."""
    cli.user()
    click.echo(
        "You're logged in as "
        + click.style(cli.data['email'], fg='cyan')
        + '.'
    )



# -*- coding: utf-8 -*-
from story.commands.login import login


def test_login(runner, init_sample_app_in_cwd):
    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()
        result = runner.run(login, exit_code=0)
        assert 'me@me.com' in result.stdout  # me@me.com is automatically set.



# -*- coding: utf-8 -*-
from blindspin import spinner

import click

import emoji

from .. import api
from .. import cli
from .. import options


@cli.cli.group()
def maintenance():
    """Manage the availability of your apps."""
    pass


@maintenance.command()
@options.app()
def check(app):
    """Displays current maintenance status."""
    cli.user()
    click.echo(f'Fetching maintenance mode for {app}… ', nl=False)
    with spinner():
        enabled = api.Apps.maintenance(app=app, maintenance=None)
    if enabled:
        click.echo(
            click.style('ON. Application is disabled.', bold=True, fg='red')
        )
    else:
        click.echo(
            click.style('off. Application is running.', bold=True, fg='green')
        )


@maintenance.command()
@options.app()
def on(app):
    """Turns maintenance–mode on."""

    cli.user()
    click.echo(f'Enabling maintenance mode for app {app}… ', nl=False)
    with spinner():
        api.Apps.maintenance(app=app, maintenance=True)
    click.echo(
        '\b' + click.style(emoji.emojize(':heavy_check_mark:'), fg='green')
    )
    click.echo(
        click.style('Application is now in maintenance–mode.', dim=True)
    )


@maintenance.command()
@options.app()
def off(app):
    """Turns maintenance–mode off."""

    cli.user()
    click.echo(f'Disabling maintenance mode for app {app}… ', nl=False)
    with spinner():
        api.Apps.maintenance(app=app, maintenance=False)
    click.echo(
        click.style('\b' + emoji.emojize(':heavy_check_mark:'), fg='green')
    )
    click.echo(click.style('Application is now running.', dim=True))



# -*- coding: utf-8 -*-
import os
import sys

from blindspin import spinner

import click

import emoji

from .. import api
from .. import awesome
from .. import cli
from .. import options
from ..helpers import datetime


def maintenance(enabled: bool) -> str:
    if enabled:
        return 'in maintenance'
    else:
        return 'running'


@cli.cli.group()
def apps():
    """Create, list, and manage apps on Storyscript Cloud."""
    pass


@apps.command(name='list')
def list_command():
    """List apps that you have access to."""
    from texttable import Texttable

    cli.user()

    with spinner():
        res = api.Apps.list()

    count = 0
    # Heads up! Texttable does not like colours.
    # So don't use click.style here.
    table = Texttable(max_width=800)
    table.set_deco(Texttable.HEADER)
    table.set_cols_align(['l', 'l', 'l'])
    all_apps = [['NAME', 'STATE', 'CREATED']]
    for app in res:
        count += 1
        date = datetime.parse_psql_date_str(app['timestamp'])
        all_apps.append(
            [
                app['name'],
                maintenance(app['maintenance']),
                datetime.reltime(date)
            ]
        )

    table.add_rows(rows=all_apps)

    if count == 0:
        click.echo('No application found. Create your first app with')
        click.echo(click.style('$ story apps create', fg='magenta'))
    else:
        click.echo(table.draw())


@apps.command()
@click.argument('name', nargs=1, required=False)
@click.option(
    '--team', type=str, help='Team name that owns this new Application'
)
def create(name, team):
    """Create a new app."""

    cli.user()
    story_yaml = cli.find_story_yml()

    if story_yaml is not None:
        click.echo(
            click.style(
                'There appears to be an Storyscript Cloud project in '
                f'{story_yaml} already.\n',
                fg='red',
            )
        )
        click.echo(
            click.style(
                'Are you trying to deploy? ' 'Try the following:', fg='red'
            )
        )
        click.echo(click.style('$ story deploy', fg='magenta'))
        sys.exit(1)

    # _is_git_repo_good()

    name = name or awesome.new()

    # Sanity check.
    if len(name) < 4:
        click.echo(
            click.style(
                'The name you specified is too short. \n'
                'Please use at least 4 characters in your app name.',
                fg='red',
            )
        )
        sys.exit(1)

    click.echo('Creating application… ', nl=False)

    with spinner():
        api.Apps.create(name=name, team=team)

    click.echo(
        '\b' + click.style(emoji.emojize(':heavy_check_mark:'), fg='green')
    )

    # click.echo('Adding git-remote... ', nl=False)
    # cli.run(f'git remote add asyncy https://git.asyncy.com/{name}')
    # click.echo(click.style(emoji.emojize(':heavy_check_mark:'), fg='green'))

    click.echo('Creating story.yml… ', nl=False)
    cli.settings_set(f'app_name: {name}\n', 'story.yml')

    click.echo(
        '\b' + click.style(emoji.emojize(':heavy_check_mark:'), fg='green')
    )

    click.echo('\nApp Name: ' + click.style(name, bold=True))
    click.echo(
        'App URL: '
        + click.style(f'https://{name}.storyscriptapp.com/', fg='blue')
        + '\n'
    )

    click.echo(
        'You are now ready to write your first Storyscript! '
        f'{emoji.emojize(":party_popper:")}'
    )
    click.echo()

    cli.track('App Created', {'App name': name})

    click.echo(' - [ ] Write a Story:')
    click.echo(
        '       $ '
        + click.style('story write http > http.story', fg='magenta')
    )
    click.echo()
    click.echo(' - [ ] Deploy to Storyscript Cloud:')
    click.echo('       $ ' + click.style('story deploy', fg='magenta'))
    click.echo()
    click.echo('We hope you enjoy your deployment experience!')


@apps.command()
@options.app()
def url(app):
    """Display the full URL of an app.

    Great to use with $(story apps url) in bash.
    """
    cli.user()
    print_nl = False

    if _isatty():
        print_nl = True

    click.echo(f'https://{app}.storyscriptapp.com/', nl=print_nl)


def _isatty():
    os.isatty(sys.stdout.fileno())


@apps.command('open')
@options.app()
def do_open(app):
    """Open the full URL of an app, in the browser."""
    cli.user()
    url = f'https://{app}.storyscriptapp.com/'

    click.echo(url)
    click.launch(url)
    sys.exit(0)


@apps.command()
@options.app()
@click.option(
    '--yes', '-y', is_flag=True, help='Assume yes to destruction confirmation.'
)
@click.option(
    '--all', is_flag=True, help='Destroy all Storyscript Cloud apps.'
)
def destroy(yes=False, app=None, all=False):
    """Destroy an app."""

    cli.user()

    if all:
        click.echo('Destroying all Storyscript Cloud applications:', err=True)
        apps = [app['name'] for app in api.Apps.list()]
    else:
        apps = [app]

    for app in apps:

        if yes or click.confirm(
            f'Do you want to destroy {app!r}?', abort=True
        ):
            click.echo(f'Destroying application {app!r}… ', nl=False)

            with spinner():
                api.Apps.destroy(app=app)
                cli.track('App Destroyed', {'App name': app})

            click.echo(
                '\b'
                + click.style(emoji.emojize(':heavy_check_mark:'), fg='green')
            )
            click.echo()

    sys.exit(0)



# -*- coding: utf-8 -*-
import pkgutil

from pytest import mark

from story import cli
from story.commands.write import write


def test_write_choices(runner):
    with runner.runner.isolated_filesystem():
        result = runner.run(write, exit_code=1)
    assert 'Please specify a template' in result.output
    assert 'Run $ story write :template_name:' in result.output


@mark.parametrize('story_name', ['http', 'function', 'if',
                                 'loop', 'slack-bot'])
@mark.parametrize('write_to_file', [True, False])
@mark.parametrize('app_name', [None, 'my_app'])
def test_write_a_story(runner, story_name, patch, write_to_file, app_name):
    patch.object(cli, 'track')
    patch.object(cli, 'get_app_name_from_yml', return_value=app_name)

    file_content = None

    with runner.runner.isolated_filesystem():
        if write_to_file:
            result = runner.run(write, exit_code=0,
                                args=[story_name, 'a.story'])

            with open('a.story', 'r') as f:
                file_content = f.read()

        else:
            result = runner.run(write, exit_code=0, args=[story_name])

    data = pkgutil.get_data('story', f'stories/{story_name}.story')

    expected_output = data.decode('utf-8') + '\n'

    if write_to_file:
        expected_output = '$ cat a.story\n' + expected_output

        assert file_content == data.decode('utf-8')

    assert expected_output == result.output

    cli.track.assert_called_with('App Bootstrapped',
                                 {
                                     'App name': app_name or 'Not created yet',
                                     'Template used': story_name
                                 })



# -*- coding: utf-8 -*-
import math
from datetime import datetime

import pytz


def parse_psql_date_str(ts: str):
    # Replace the ":" in the timezone field for datetime.
    datetime_ts = ts[0 : ts.rindex(':')] + ts[ts.rindex(':') + 1 :]
    return datetime.strptime(
        datetime_ts, '%Y-%m-%dT%H:%M:%S.%f%z'
    ).astimezone()


# https://gist.github.com/deontologician/3503910
def reltime(date, compare_to=None, at='@'):
    """
    Takes a datetime and returns a relative representation of the
    time.
    :param date: The date to render relatively
    :param compare_to: what to compare the date to. Defaults to datetime.now()
    :param at: date/time separator. defaults to "@". "at" is also reasonable.
    >>> from datetime import datetime, timedelta
    >>> today = datetime(2050, 9, 2, 15, 00)
    >>> earlier = datetime(2050, 9, 2, 12)
    >>> reltime(earlier, today)
    'today @ 12pm'
    >>> yesterday = today - timedelta(1)
    >>> reltime(yesterday, compare_to=today)
    'yesterday @ 3pm'
    >>> reltime(datetime(2050, 9, 1, 15, 32), today)
    'yesterday @ 3:32pm'
    >>> reltime(datetime(2050, 8, 31, 16), today)
    'Wednesday @ 4pm (2 days ago)'
    >>> reltime(datetime(2050, 8, 26, 14), today)
    'last Friday @ 2pm (7 days ago)'
    >>> reltime(datetime(2049, 9, 2, 12, 00), today)
    'September 2nd, 2049 @ 12pm (last year)'
    >>> today = datetime(2012, 8, 29, 13, 52)
    >>> last_mon = datetime(2012, 8, 20, 15, 40, 55)
    >>> reltime(last_mon, today)
    'last Monday @ 3:40pm (9 days ago)'
    """

    def ordinal(n):
        """
        Returns a string ordinal representation of a number
        Taken from: http://stackoverflow.com/a/739301/180718
        """
        if 10 <= n % 100 < 20:
            return str(n) + 'th'
        else:
            return str(n) + {1: 'st', 2: 'nd', 3: 'rd'}.get(n % 10, 'th')

    compare_to = compare_to or pytz.UTC.localize(datetime.now())
    if date > compare_to:
        return NotImplementedError('reltime only handles dates in the past')
    # get timediff values
    diff = compare_to - date
    if diff.seconds < 60 * 60 * 8:  # less than a business day?
        days_ago = diff.days
    else:
        days_ago = diff.days + 1
    months_ago = compare_to.month - date.month
    years_ago = compare_to.year - date.year
    weeks_ago = int(math.ceil(days_ago / 7.0))
    # get a non-zero padded 12-hour hour
    hr = date.strftime('%I')
    if hr.startswith('0'):
        hr = hr[1:]
    wd = compare_to.weekday()
    # calculate the time string
    if date.minute == 0:
        time = '{0}{1}'.format(hr, date.strftime('%p').lower())
    else:
        time = '{0}:{1}'.format(hr, date.strftime('%M%p').lower())
    # calculate the date string
    if days_ago == 0:
        datestr = 'today {at} {time}'
    elif days_ago == 1:
        datestr = 'yesterday {at} {time}'
    elif (
        wd in (5, 6) and days_ago in (wd + 1, wd + 2)
    ) or wd + 3 <= days_ago <= wd + 8:
        # this was determined by making a table of wd versus days_ago and
        # divining a relationship based on everyday speech. This is somewhat
        # subjective I guess!
        datestr = 'last {weekday} {at} {time} ({days_ago} days ago)'
    elif days_ago <= wd + 2:
        datestr = '{weekday} {at} {time} ({days_ago} days ago)'
    elif years_ago == 1:
        datestr = '{month} {day}, {year} {at} {time} (last year)'
    elif years_ago > 1:
        datestr = '{month} {day}, {year} {at} {time} ({years_ago} years ago)'
    elif months_ago == 1:
        datestr = '{month} {day} {at} {time} (last month)'
    elif months_ago > 1:
        datestr = '{month} {day} {at} {time} ({months_ago} months ago)'
    else:
        # not last week, but not last month either
        datestr = '{month} {day} {at} {time} ({days_ago} days ago)'
    return datestr.format(
        time=time,
        weekday=date.strftime('%A'),
        day=ordinal(date.day),
        days=diff.days,
        days_ago=days_ago,
        month=date.strftime('%B'),
        years_ago=years_ago,
        months_ago=months_ago,
        weeks_ago=weeks_ago,
        year=date.year,
        at=at,
    )



# -*- coding: utf-8 -*-
import json
import re
from urllib.error import URLError

import click

from pytest import fixture, mark

import requests

from story import api
from story.environment import SS_GRAPHQL


@fixture()
def patch_graphql(patch):
    patch.object(api, 'graphql')
    patch.object(api.Apps, 'get_uuid_from_hostname')


@mark.parametrize('with_api_errors', [True, False])
@mark.parametrize('api_throws_an_exception',
                  [
                      None,
                      URLError('my_reason'),
                      requests.RequestException(),
                      KeyboardInterrupt()
                  ])
def test_graphql(runner, patch, with_api_errors, api_throws_an_exception):
    if api_throws_an_exception:
        patch.object(click, 'get_current_context')
        patch.object(requests, 'post', side_effect=api_throws_an_exception)
    else:
        patch.object(requests, 'post')

    if with_api_errors:
        requests.post.return_value.json.return_value = {
            'errors': [
                {
                    'message': 'my_error_message'
                }
            ]
        }

        patch.object(click, 'get_current_context')

    my_query = {'my_query': 'hello_world'}
    variables = {
        'var1': 'val1',
        'var2': 'val2'
    }

    with runner.runner.isolated_filesystem():
        from story import cli

    patch.object(cli, 'get_access_token')

    ret_val = api.graphql(my_query, **variables)

    requests.post.assert_called_with(
        SS_GRAPHQL,
        data=json.dumps({
            'query': my_query,
            'variables': variables
        }),
        headers={
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': f'Bearer {cli.get_access_token()}',
        },
        timeout=10
    )

    if with_api_errors or api_throws_an_exception:
        click.get_current_context().exit.assert_called_with(1)
        assert ret_val is None
    else:
        assert ret_val == requests.post.return_value.json.return_value


app_name = 'my_app'


@fixture()
def assert_simple_graphql_call(patch_graphql, magic):
    def wrapper(method, method_args, query, assert_hostname_call=True,
                args={}, graphql_response=None):

        if graphql_response is not None:
            api.graphql.return_value = graphql_response

        ret_val = method(*method_args)
        actual_query = api.graphql.mock_calls[0][1][0]

        if assert_hostname_call:
            args['app'] = api.Apps.get_uuid_from_hostname.return_value

        assert sanitise_graphql_query(query) == sanitise_graphql_query(
            actual_query)

        api.graphql.assert_called_with(
            actual_query,
            **args
        )

        if assert_hostname_call:
            api.Apps.get_uuid_from_hostname.assert_called_with(app_name)

        return ret_val

    return wrapper


@mark.parametrize('graphql_returned_nothing', [True, False])
def test_config_get(assert_simple_graphql_call, graphql_returned_nothing):
    expected_query = """
    query($app: UUID!){
      allReleases(condition: {appUuid: $app},
                  first: 1, orderBy: ID_DESC){
        nodes{
          config
        }
      }
    }
    """

    mock_res = None

    if graphql_returned_nothing:
        mock_res = {}

    ret_val = assert_simple_graphql_call(method=api.Config.get,
                                         method_args=[app_name],
                                         query=expected_query,
                                         graphql_response=mock_res)

    if graphql_returned_nothing:
        assert ret_val == {}
    else:
        assert ret_val == api.graphql()['data']['allReleases']['nodes'][0][
            'config']


def test_apps_destroy(assert_simple_graphql_call):
    expected_query = """
    mutation ($data: UpdateAppByUuidInput!){
              updateAppByUuid(input: $data){
                app{
                  uuid
                }
              }
            }
    """

    ret_val = assert_simple_graphql_call(
        method=api.Apps.destroy,
        method_args=[app_name],
        assert_hostname_call=False,
        args={
            'data': {
                'uuid': api.Apps.get_uuid_from_hostname(
                    app_name),
                'appPatch': {'deleted': True}
            }
        },
        query=expected_query)
    api.Apps.get_uuid_from_hostname.assert_called_with(app_name)
    assert ret_val is None


def test_apps_list(assert_simple_graphql_call):
    expected_query = """
    query{
              allApps(condition: {deleted: false}, orderBy: NAME_ASC){
                nodes{
                  name
                  timestamp
                  maintenance
                }
              }
            }
    """

    ret_val = assert_simple_graphql_call(
        method=api.Apps.list,
        method_args=[],
        assert_hostname_call=False,
        query=expected_query)

    assert ret_val == api.graphql()['data']['allApps']['nodes']


def sanitise_graphql_query(query: str) -> str:
    query = query.strip()
    query = re.sub(r'[\s]{2,}', ' ', query)
    query = query.replace(' ', '_space_')
    query = query.replace('\n', '_new_line_')
    return query



# -*- coding: utf-8 -*-

import json
import os
import subprocess
import sys
import time
import typing
from urllib.parse import urlencode
from uuid import uuid4

from blindspin import spinner

import click

import click_help_colors


import emoji

from raven import Client

from requests import Session

from . import storage, utils
from .ensure import ensure_latest
from .helpers.didyoumean import DYMGroup
from .support import echo_support
from .version import compiler_version
from .version import version as story_version

# Initiate requests session, for connection pooling.
requests = Session()

# Typing hints.
Content = typing.Union[str, typing.Mapping, typing.List]

if not os.getenv('TOXENV'):
    enable_reporting = True
    sentry = Client(
        'https://007e7d135737487f97f5fe87d5d85b55@sentry.io/1206504'
    )
else:
    enable_reporting = False
    sentry = Client()

data = None


def get_access_token():
    return data['access_token']


def get_user_id():
    """Returns the current user id, if any is available."""
    return data.get('id')


def track_profile():
    _make_tracking_http_request(
        'https://stories.storyscriptapp.com/track/profile',
        {
            'id': str(get_user_id()),
            'profile': {
                'Name': data['name'],
                'Email': data.get('email'),
                'GitHub Username': data.get('username'),
                'Timezone': time.tzname[time.daylight],
            },
        },
    )


def _make_tracking_http_request(url, json_data):
    """
    Forks the HTTP call into the background.
    The caller does not make the call, and this method returns just after
    forking the process.

    The child makes the call, and does not return back to the caller,
    but instead just exits (sys.exit(0)). This is to ensure that whatever
    needs to happen after this method is called, does not happen twice.
    """
    if not enable_reporting:
        return
    if sys.platform != 'linux' and sys.platform != 'darwin':
        try:
            requests.post(url, json=json_data)
        except Exception:
            # ignore issues with tracking
            pass

        return

    pid = os.fork()
    if pid > 0:
        # This is the parent process.
        return

    try:
        requests.post(url, json=json_data)
    except Exception:
        # ignore issues with tracking
        pass

    # This is the child process. Exit now.
    sys.exit(0)


def track(event_name, extra: dict = None):
    if extra is None:
        extra = {}

    extra['CLI version'] = story_version
    _make_tracking_http_request(
        'https://stories.storyscriptapp.com/track/event',
        {
            'id': str(get_user_id()),
            'event_name': event_name,
            'event_props': extra,
        },
    )


class NoCredsException(BaseException):
    pass


def initiate_login():
    global data

    click.echo(
        'Hi! Thank you for using '
        + click.style('Storyscript Cloud', fg='magenta')
        + '.'
    )
    click.echo('Please login with GitHub to get started.')

    state = uuid4()

    query = {'state': state}

    url = f'https://stories.storyscriptapp.com/github?{urlencode(query)}'

    click.launch(url)
    click.echo()
    click.echo("Visit this link if your browser doesn't open automatically:")
    click.echo(url)
    click.echo()

    while True:
        with spinner():
            try:
                url = (
                    'https://stories.storyscriptapp.com'
                    '/github/oauth_callback'
                )
                r = requests.get(url=url, params={'state': state})

                if r.text == 'null':
                    raise NoCredsException()

                r.raise_for_status()
                break

            except IOError as e:
                click.echo(f'The following network error occurred:', err=True)
                click.echo(f'\t{str(e)}', err=True)
                click.echo('The Storyscript Cloud CLI '
                           'will retry automatically...')
                click.echo()
                time.sleep(2)
            except NoCredsException:
                time.sleep(0.5)
            except KeyboardInterrupt:
                click.echo('Login failed. Please try again.')
                sys.exit(1)

    if r.json().get('beta') is False:
        click.echo('Hello! Storyscript Cloud is in private beta at this time.')
        click.echo(
            'We\'ve added you to our beta testers queue, '
            'and you should hear from us\nshortly via email'
            ' (which is linked to your GitHub account).'
        )
        sys.exit(1)

    data = r.json()
    for k, v in data.items():
        storage.config.store(k, v)

    init()

    click.echo(emoji.emojize(':waving_hand:') + f'  Welcome {data["name"]}!')
    click.echo()
    click.echo('Create a new app with:')
    print_command('story apps create')

    click.echo()

    click.echo('To list all your apps:')
    print_command('story apps')

    click.echo()
    track('Login Completed')
    track_profile()


def user() -> dict:
    """Get the active user."""

    global data

    if data:
        return data
    else:
        initiate_login()
        return data


def print_command(command):
    """Prints a command to the CLI."""

    click.echo('  $ ' + click.style(command, fg='magenta'))


def find_story_yml():
    return utils.find_story_yml()


def get_app_name_from_yml():
    return utils.get_app_name_from_yml()


def get_asyncy_yaml():
    return utils.get_asyncy_yaml()


def assert_project(command, app, default_app, allow_option):
    if app is None:
        click.echo(
            click.style('No Storyscript Cloud application found.', fg='red')
        )
        click.echo()
        click.echo('Create an application with:')

        print_command('story apps create')

        sys.exit(1)

    elif not allow_option and app != default_app:
        click.echo(
            click.style(
                'The --app option is not allowed with the {} command.'.format(
                    command
                ),
                fg='red',
            )
        )
        sys.exit(1)
    return app


def settings_set(content: Content, location: str):
    """Overwrites settings, to given location and content."""

    # Ensure the path is created and exists..
    loc_dir = os.path.abspath(location)
    loc_dir = os.path.dirname(loc_dir)

    os.makedirs(loc_dir, exist_ok=True)

    # If content is an object-like...
    if isinstance(content, (list, dict)):
        content = json.dumps(content, indent=2)

    # Write to the file.
    with open(location, 'w+') as file:
        file.write(content)


def init(config_path=None):
    global data

    storage.init_storage(config_path)

    data = storage.config.as_dict()

    try:
        sentry.user_context({'id': get_user_id(), 'email': data['email']})
    except Exception:
        pass


def stream(cmd: str):
    process = subprocess.Popen(cmd.split(' '), stdout=subprocess.PIPE)

    while True:

        output = process.stdout.readline()

        if output == b'' and process.poll() is not None:
            break

        if output:
            click.echo(output.strip())


def echo_version(*args, **kwargs):
    click.echo(
        click.style('Storyscript CLI', fg='magenta')
        + ': v'
        + story_version
        + click.style(', ', dim=True)
        + click.style('Storyscript Compiler', fg='cyan')
        + ': v'
        + compiler_version
        + '.'
    )


class CLIGroup(DYMGroup, click_help_colors.HelpColorsGroup):
    pass


# @click.option('--url', callback=version)
@click.group(
    cls=CLIGroup,
    help_headers_color='yellow',
    help_options_color='magenta',
    add_help_option=True,
    no_args_is_help=True,
    invoke_without_command=True,
    context_settings={
        'help_option_names': ['-h', '--help'],
        'auto_envvar_prefix': 'STORY',
    },
)
@click.option(
    '--version',
    'do_version',
    is_flag=True,
    help='Show version information and exit',
)
@click.option('--config', 'do_config', is_flag=True, hidden=True)
@click.option('--config_path', 'config_path', hidden=True)
@click.option('--cache', 'do_cache', is_flag=True, hidden=True)
@click.option(
    '--disable-version-check', 'dont_check', is_flag=True, hidden=True
)
@click.option('--completion', 'do_completion', is_flag=True, hidden=True)
@click.option('--reset', 'do_reset', is_flag=True, hidden=True)
@click.option('--support', 'do_support', is_flag=True, hidden=True)
def cli(
    app=None,
    do_version=False,
    do_config=False,
    do_cache=False,
    do_reset=False,
    do_support=False,
    do_completion=False,
    dont_check=False,
    config_path=False,
):
    """
    Hello! Welcome to Storyscript.

    We hope you enjoy, & we look forward to your feedback!

    Documentation: https://docs.storyscript.io/
    """

    if do_version:
        echo_version()
        sys.exit(0)

    elif do_cache:
        click.echo(storage.cache.path)
        sys.exit(0)

    elif do_config:
        click.echo(storage.config.path)
        sys.exit(0)

    elif do_reset:
        reset()
        sys.exit(0)
    elif do_support:
        echo_support()

    else:
        init(config_path=config_path)

        # Check for new versions, if allowed.
        if not dont_check:
            ensure_latest()


def reset():
    storage.cache.remove_file_on_disk()
    storage.config.remove_file_on_disk()
    click.echo('Storyscript CLI installation reset.\n'
               'Run the following command to login again:')
    print_command('story login')
